from copy import deepcopy
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.restapi.testing import RelativeSession

import pytest


class TestServicesSite:
    @pytest.fixture(autouse=True)
    def _api_session(self, app):
        url = app.absolute_url()
        api_session = RelativeSession(url)
        api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        api_session.headers.update({"Accept": "application/json"})
        self.api_session = api_session


class TestServicesSitesGET(TestServicesSite):
    def test_sites_get_anonymous(self, app):
        self.api_session.auth = None
        response = self.api_session.get("@sites")
        assert response.status_code == 200
        data = response.json()
        assert isinstance(data, dict)
        assert data["can_manage"] is False

    def test_sites_get_authenticated(self, app):
        response = self.api_session.get("@sites")
        assert response.status_code == 200
        data = response.json()
        assert isinstance(data, dict)
        assert data["can_manage"] is True

    def test_sites_get_distributions(self, app):
        response = self.api_session.get("@sites")
        data = response.json()
        distributions = data["distributions"]
        assert isinstance(distributions, list)
        # We expect our testing distribution, but there may be a volto
        # distribution as well.
        assert len(distributions) >= 1
        assert "testing" in [distro["name"] for distro in distributions]
        for distro in distributions:
            if distro["name"] == "testing":
                assert distro["title"] == "Testing Plone Site"

    def test_sites_get_sites(self, app):
        response = self.api_session.get("@sites")
        data = response.json()
        sites = data["sites"]
        assert isinstance(sites, list)
        assert len(sites) == 1
        assert sites[0]["id"] == "plone"
        assert sites[0]["needs_upgrade"] is False


class TestServicesSitesPOST(TestServicesSite):
    @pytest.fixture(autouse=True)
    def _answers(self):
        self.answers = {
            "site_id": "PloneBrasil",
            "title": "Plone site in Brazil",
            "description": "A simple Plone Site",
            "default_language": "pt-br",
            "portal_timezone": "America/Sao_Paulo",
            "setup_content": True,
        }

    @pytest.mark.parametrize(
        "distribution_name",
        [
            "testing",
        ],
    )
    def test_sites_create_success(self, app, distribution_name):
        response = self.api_session.post(
            f"@sites/{distribution_name}", json=self.answers
        )
        assert response.status_code == 200
        data = response.json()
        assert data["id"] == "PloneBrasil"
        assert data["@type"] == "Plone Site"

    @pytest.mark.parametrize(
        "distribution_name",
        [
            "someStrangeName",
            "404",
        ],
    )
    def test_sites_create_failure(self, app, distribution_name):
        """Test POST with non-existing distribution names."""
        response = self.api_session.post(
            f"@sites/{distribution_name}", json=self.answers
        )
        assert response.status_code == 404
        data = response.json()
        assert data["message"].startswith("Resource not found:")

    def test_sites_create_empty_site_id(self, app):
        distribution_name = "testing"
        # Empty site_id
        answers = {"site_id": "", "title": "Site"}

        response = self.api_session.post(f"@sites/{distribution_name}", json=answers)
        assert response.status_code == 400
        data = response.json()
        assert data["message"].startswith("Invalid data for site creation")

    def test_sites_create_empty_title(self, app):
        distribution_name = "testing"
        # Empty title
        answers = {"site_id": "Site", "title": ""}

        response = self.api_session.post(f"@sites/{distribution_name}", json=answers)
        assert response.status_code == 400
        data = response.json()
        assert data["message"].startswith("Invalid data for site creation")

    def test_sites_create_invalid_language(self, app):
        distribution_name = "testing"
        answers = deepcopy(self.answers)
        answers["default_language"] = "klingon"

        response = self.api_session.post(f"@sites/{distribution_name}", json=answers)
        assert response.status_code == 400
        data = response.json()
        assert data["message"].startswith("Invalid data for site creation")

    def test_sites_create_invalid_timezone(self, app):
        distribution_name = "testing"
        answers = deepcopy(self.answers)
        answers["portal_timezone"] = "something"

        response = self.api_session.post(f"@sites/{distribution_name}", json=answers)
        assert response.status_code == 400
        data = response.json()
        assert data["message"].startswith("Invalid data for site creation")



"""Plone Distribution support."""

from pathlib import Path

import logging


DEFAULT_PATH = Path(__file__).parent / "default"
PACKAGE_NAME = "plone.distribution"

logger = logging.getLogger(PACKAGE_NAME)



from pkg_resources import get_distribution

import sys


project = "plone.distribution"
copyright = "2023, Plone Foundation"

# The full version, including alpha/beta/rc tags.
release = get_distribution(project).version
# The short X.Y version.
version = ".".join(release.split(".")[0:2])

# The suffix of source filenames.
source_suffix = {
    ".md": "markdown",
    ".rst": "restructuredtext",
}

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinxext.opengraph",
    "sphinx.ext.graphviz",
]
master_doc = "index"

# locale_dirs = ['translated/']
language = "en"

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual])
# This enables PDF generation.
latex_documents = [
    (
        "index",
        "ploneapi.tex",
        "plone.distribution Documentation",
        "",
        "manual",
    )
]


class Mock:
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    @classmethod
    def __getattr__(cls, name):
        if name in ("__file__", "__path__"):
            return "/dev/null"
        elif name[0] == name[0].upper():
            mockType = type(name, (), {})
            mockType.__module__ = __name__
            return mockType
        else:
            return Mock()


MOCK_MODULES = ["lxml"]
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()


# -- Options for myST markdown conversion to html -----------------------------

# For more information see:
# https://myst-parser.readthedocs.io/en/latest/syntax/optional.html
myst_enable_extensions = [
    "deflist",  # You will be able to utilise definition lists
    # https://myst-parser.readthedocs.io/en/latest/syntax/optional.html#definition-lists
    "colon_fence",  # You can also use ::: delimiters to denote code fences,\
    #  instead of ```.
    "substitution",  # plone.restapi \
    # https://myst-parser.readthedocs.io/en/latest/syntax/optional.html#substitutions-with-jinja2
    "html_image",  # For inline images. \
    # See https://myst-parser.readthedocs.io/en/latest/syntax/optional.html
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
html_logo = "_static/logo.svg"
html_favicon = "_static/favicon.ico"
html_css_files = ["styles.css", ("print.css", {"media": "print"})]
html_static_path = [
    "_static",  # Last path wins. See https://github.com/plone/documentation/pull/1442
]



from io import BytesIO
from plone.distribution.utils import request as utils_request
from zope.component import provideAdapter
from zope.i18n.interfaces import IUserPreferredLanguages
from zope.publisher.browser import BrowserLanguages
from zope.publisher.interfaces.http import IHTTPRequest
from ZPublisher.HTTPRequest import HTTPRequest

import pytest


class TestUtilsRequest:
    @pytest.mark.parametrize(
        "language,expected",
        [
            ("en-US,en;q=0.5", "en-us"),
            ("pt-BR,pt;q=0.5", "pt-br"),
            ("pt-BO,pt;q=0.5", "pt"),
            ("pt-PT,pt;q=0.5", "pt"),
            ("en;q=0.5", "en"),
            ("es;q=0.5", "es"),
            ("de;q=0.5", "de"),
        ],
    )
    def test_extract_browser_language(self, language, expected):
        provideAdapter(BrowserLanguages, [IHTTPRequest], IUserPreferredLanguages)
        request = HTTPRequest(
            BytesIO(),
            {
                "HTTP_ACCEPT_LANGUAGE": language,
                "SERVER_NAME": "nohost",
                "SERVER_PORT": "8080",
            },
            None,
        )
        value = utils_request.extract_browser_language(request)
        assert value == expected



from plone.distribution.testing import FUNCTIONAL_TESTING
from plone.distribution.testing import INTEGRATION_TESTING
from pytest_plone import fixtures_factory


pytest_plugins = ["pytest_plone"]


globals().update(
    fixtures_factory(
        (
            (FUNCTIONAL_TESTING, "functional"),
            (INTEGRATION_TESTING, "integration"),
        )
    )
)



from plone.distribution.utils import data as utils_data

import pytest


class TestUtilsStaConvertDataURI:
    @property
    def func(self):
        return utils_data.convert_data_uri_to_b64

    @pytest.mark.parametrize(
        "value,expected",
        [
            [
                "name=teste;data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg==",  # noQA
                b"filenameb64:dGVzdGU7ZGF0YTppbWFnZS9wbmc=;datab64:iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg==",  # noQA
            ],
            ["", b""],
        ],
    )
    def test_check_conversion(self, value: str, expected: bytes):
        result = self.func(value)
        assert result == expected



from pathlib import Path

import json
import pytest


@pytest.fixture
def data_path() -> Path:
    cur_dir = Path(__file__).parent.resolve()
    return cur_dir / "data"


@pytest.fixture
def language_schema(data_path):
    return json.loads((data_path / "with_default_language.json").read_text())


@pytest.fixture
def valid_schema(data_path):
    return json.loads((data_path / "valid.json").read_text())


@pytest.fixture
def invalid_schema(data_path):
    return json.loads((data_path / "invalid.json").read_text())



from plone.distribution.utils import schema as utils_schema

import pytest


class TestUtilsSchemaValidateJsonSchema:
    @property
    def func(self):
        return utils_schema.validate_jsonschema

    def test_validate_jsonschema_success(self, app, valid_schema):
        schema = valid_schema["schema"]
        result = self.func(schema)
        assert result is True

    def test_validate_jsonschema_fail(self, app, invalid_schema):
        schema = invalid_schema["schema"]
        result = self.func(schema)
        assert result is False

    def test_validate_jsonschema_strict_fail(self, app, valid_schema):
        schema = valid_schema["schema"]
        # Remove default_language property (that is only recommended, not required)
        del schema["properties"]["default_language"]
        result = self.func(schema, strict=True)
        assert result is False


class TestUtilsSchemaEnrichJsonSchema:
    @property
    def func(self):
        return utils_schema.enrich_jsonschema

    @pytest.mark.parametrize("key", ["languages", "timezones"])
    def test_enrich_jsonschema(self, app, key):
        schema = {}
        result = self.func(schema)
        assert "definitions" in result
        definitions = result["definitions"]
        assert key in definitions


class TestUtilsSchemaEnrichUiSchema:
    @property
    def func(self):
        return utils_schema.enrich_uischema

    def test_enrich_uischema(self, app, valid_schema):
        jsonschema = valid_schema["schema"]
        uischema = valid_schema["uischema"]
        assert len(uischema) == 0
        result = self.func(uischema, jsonschema)
        assert result["ui:order"][0] == "site_id"
        assert result["ui:order"][1] == "title"
        assert result["ui:order"][2] == "description"


class TestUtilsSchemaProcessRawSchema:
    @property
    def func(self):
        return utils_schema.process_raw_schema

    def test_enrich_uischema(self, app, valid_schema):
        result = self.func(valid_schema)
        assert "schema" in result
        assert "uischema" in result
        uischema = result["uischema"]
        assert "ui:order" in uischema


class TestUtilsSchemaDefaultLanguage:
    def process_schema(self, raw_schema):
        schema = utils_schema.process_raw_schema(raw_schema)
        return schema["uischema"], schema["schema"]

    @property
    def func(self):
        return utils_schema.should_provide_default_language_default

    def test_should_provide_default_language_default_true(self, app, valid_schema):
        uischema, jsonschema = self.process_schema(valid_schema)
        result = self.func(uischema, jsonschema)
        assert result is True

    def test_should_provide_default_language_default_false(self, app, language_schema):
        uischema, jsonschema = self.process_schema(language_schema)
        result = self.func(uischema, jsonschema)
        assert result is False



"""Service to create a Plone Site."""

from plone.distribution import logger
from plone.distribution.api import distribution as dist_api
from plone.distribution.api import site as site_api
from plone.distribution.utils.validation import validate_answers
from plone.restapi.deserializer import json_body
from plone.restapi.services import Service
from zExceptions import BadRequest
from zExceptions import NotFound
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import plone.protect.interfaces


@implementer(IPublishTraverse)
class SiteCreate(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Consume any path segments after /@sites as parameters
        self.params.append(name)
        return self

    def reply(self):
        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)
        data = json_body(self.request)
        distribution_name = (
            self.params[0] if self.params else data.get("distribution", "default")
        )
        self.errors = []
        # Get distribution
        try:
            distribution = dist_api.get(distribution_name)
        except ValueError:
            raise NotFound(f"No distribution named {distribution_name}.")

        # Validate answers
        if not validate_answers(answers=data, schema=distribution.schema):
            raise BadRequest("Invalid data for site creation.")

        try:
            site = site_api.create(
                self.context,
                distribution_name=distribution_name,
                answers=data,
            )
        except KeyError:
            logger.error("Error creating the site.", exc_info=True)
            raise BadRequest("Error creating the site.")
        site_info = {
            "@id": site.absolute_url(),
            "@type": site.portal_type,
            "id": site.id,
            "title": site.title,
            "description": site.description,
        }
        return site_info



from plone.distribution.utils import schema as utils_schema
from plone.distribution.utils import validation

import pytest


@pytest.fixture
def schema(valid_schema) -> dict:
    return utils_schema.enrich_jsonschema(valid_schema["schema"])


@pytest.fixture
def valid_answers() -> dict:
    return {
        "site_id": "Plone",
        "title": "Testing Plone Site",
        "description": "A new Plone Site",
        "default_language": "en",
        "portal_timezone": "UTC",
        "setup_content": True,
    }


@pytest.fixture
def invalid_answers() -> dict:
    return {
        "site_id": "Plone",
        "title": 123,
        "description": "A new Plone Site",
        "default_language": "en",
        "portal_timezone": "UTC",
        "setup_content": True,
    }


class TestUtilsSchemaValidateAnswers:
    @pytest.fixture(autouse=True)
    def _init(self, app):
        self.app = app

    @property
    def func(self):
        return validation.validate_answers

    def test_valid_schema_valid_answers(self, valid_answers, schema):
        result = self.func(valid_answers, schema)
        assert result is True

    def test_valid_schema_invalid_answers(self, invalid_answers, schema):
        result = self.func(invalid_answers, schema)
        assert result is False

    def test_invalid_schema_valid_answers(self, valid_answers, invalid_schema):
        schema = invalid_schema["schema"]
        result = self.func(valid_answers, schema)
        assert result is False

    def test_invalid_schema_invalid_answers(self, invalid_answers, invalid_schema):
        schema = invalid_schema["schema"]
        result = self.func(invalid_answers, schema)
        assert result is False



from plone import api
from plone.dexterity.schema import SCHEMA_CACHE
from plone.distribution import logger
from plone.distribution.core import Distribution
from plone.distribution.utils.data import convert_data_uri_to_b64
from plone.exportimport.importers import get_importer
from Products.CMFPlone.Portal import PloneSite

import transaction


def default_pre_handler(answers: dict) -> dict:
    """Process answers before creating a new site."""
    return answers


def default_handler(
    distribution: Distribution, site: PloneSite, answers: dict
) -> PloneSite:
    """Default handler to create a new site."""
    # Process answers
    profiles = distribution.profiles
    setup_content = answers.get("setup_content", False)
    setup_tool = site["portal_setup"]
    for profile_id in profiles:
        setup_tool.runAllImportStepsFromProfile(f"profile-{profile_id}")

    # Add default content if needed
    if setup_content:
        contents = distribution.contents
        # First process any content profiles
        content_profiles = contents["profiles"]
        for profile_id in content_profiles:
            setup_tool.runAllImportStepsFromProfile(f"profile-{profile_id}")
        # Process content import from json
        content_json_path = contents["json"]
        if content_json_path:
            # If there is no savepoint most tests fail with a PosKeyError
            transaction.savepoint(optimistic=True)
            # Invalidate the schema cache to make sure we get up to date behaviors.
            # Normally this happens on commit, but we didn't commit yet.
            SCHEMA_CACHE.clear()
            importer = get_importer(site)
            importer.import_site(content_json_path)
            transaction.commit()
    return site


def post_handler(
    distribution: Distribution, site: PloneSite, answers: dict
) -> PloneSite:
    """After site creation, run last steps."""
    name = distribution.name
    raw_logo = answers.get("site_logo")
    if raw_logo:
        logo = convert_data_uri_to_b64(raw_logo)
        logger.info(f"{name}: Set logo")
        api.portal.set_registry_record("plone.site_logo", logo)
    return site



from copy import deepcopy
from plone.app.testing.interfaces import PLONE_SITE_ID
from plone.app.testing.interfaces import SITE_OWNER_NAME
from plone.app.testing.interfaces import SITE_OWNER_PASSWORD
from plone.app.testing.interfaces import TEST_USER_ID
from plone.app.testing.interfaces import TEST_USER_NAME
from plone.app.testing.interfaces import TEST_USER_PASSWORD
from plone.app.testing.interfaces import TEST_USER_ROLES
from plone.app.testing.layers import PloneFixture
from plone.testing import zope
from zope.globalrequest import setRequest


DEFAULT_DISTRIBUTION = "testing"
DEFAULT_ANSWERS = {
    "site_id": PLONE_SITE_ID,
    "title": "Plone Site",
    "description": "A Plone Site",
    "site_logo": "name=teste;data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg==",  # noQA
    "default_language": "en",
    "portal_timezone": "America/Sao_Paulo",
    "setup_content": True,
}


class PloneDistributionFixture(PloneFixture):
    """Base Fixture to test plone distributions.

    This Fixture should be extended before being used.
    """

    PACKAGE_NAME: str = ""
    SITES: tuple = ()

    _distribution_products = (
        ("plone.app.contenttypes", {"loadZCML": True}),
        ("plone.restapi", {"loadZCML": True}),
        ("plone.volto", {"loadZCML": True, "silent": True}),
        ("plone.distribution", {"loadZCML": True}),
    )

    @property
    def products(self):
        """Merge products available on PloneFixture, with ones needed here."""
        products = []
        products.extend([p for p in super().products])
        products.extend([p for p in self._distribution_products])
        if self.PACKAGE_NAME:
            products.append((self.PACKAGE_NAME, {"loadZCML": True}))
        return tuple(products)

    @property
    def sites(self):
        """Guarantee there is at least one site created."""
        sites = self.SITES
        if not sites:
            sites = ((DEFAULT_DISTRIBUTION, deepcopy(DEFAULT_ANSWERS)),)
        return sites

    def setUpDefaultContent(self, app):
        """Create a Plone site using plone.distribution."""
        from plone.distribution.api import site as site_api

        # Create the owner user and "log in" so that the site object gets
        # the right ownership information
        app["acl_users"].userFolderAddUser(
            SITE_OWNER_NAME, SITE_OWNER_PASSWORD, ["Manager"], []
        )

        setRequest(app.REQUEST)
        zope.login(app["acl_users"], SITE_OWNER_NAME)
        sites = self.sites
        for distribution_name, answers in sites:
            site_id = answers["site_id"]
            # Create Plone site
            site_api._create_site(
                context=app, distribution_name=distribution_name, answers=answers
            )

            # Create the test user. (Plone)PAS does not have an API to create a
            # user with different userid and login name, so we call the plugin
            # directly.
            pas = app[site_id]["acl_users"]
            pas.source_users.addUser(TEST_USER_ID, TEST_USER_NAME, TEST_USER_PASSWORD)
            for role in TEST_USER_ROLES:
                pas.portal_role_manager.doAssignRoleToPrincipal(TEST_USER_ID, role)

        # Log out again
        zope.logout()
        setRequest(None)



from base64 import b64encode


def convert_data_uri_to_b64(raw_data: str) -> bytes:
    """Convert data-uri format to one suitable to be used with the plone.app.registry."""
    response = b""
    if "base64," in raw_data:
        headers, data = raw_data.split("base64,")
        filename = headers.split("name=")[1][:-1]
        filenameb64 = b64encode(filename.encode("utf-8")).decode("utf-8")
        response = f"filenameb64:{filenameb64};datab64:{data}".encode()
    return response



from plone import api
from Products.CMFCore.WorkflowCore import WorkflowException
from zope.component.hooks import setSite

import pytest


@pytest.fixture()
def portal(portal_default):
    setSite(portal_default)
    return portal_default


class TestDistributionDefault:
    @pytest.mark.parametrize(
        "attr,expected",
        [
            ("title", "Testing Plone Site"),
            ("description", "A Plone Site"),
            ("exclude_from_nav", False),
        ],
    )
    def test_plone_site_attributes(self, portal, attr, expected):
        assert getattr(portal, attr) == expected

    @pytest.mark.parametrize(
        "package,expected",
        [
            ("plone.restapi", True),
            ("plone.app.contenttypes", True),
            ("plone.app.caching", True),
            ("plonetheme.barceloneta", True),
            ("plone.volto", True),
        ],
    )
    def test_dependencies_installed(self, installer, package, expected):
        assert installer.is_product_installed(package) is expected

    def test_plone_logo_applied(self, portal):
        expected_result = b"filenameb64:dGVzdGU7ZGF0YTppbWFnZS9wbmc=;datab64:iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg=="  # noQA
        assert api.portal.get_registry_record("plone.site_logo") == expected_result

    @pytest.mark.parametrize(
        "path,title,portal_type,review_state",
        [
            ("/", "Testing Plone Site", "Plone Site", ""),
        ],
    )
    def test_content_created(self, portal, path, title, portal_type, review_state):
        with api.env.adopt_roles(
            [
                "Manager",
            ]
        ):
            content = api.content.get(path=path)
        assert content.title == title
        assert content.portal_type == portal_type
        if review_state:
            assert api.content.get_state(content) == review_state
        else:
            with pytest.raises(WorkflowException) as exc:
                api.content.get_state(content)
            assert "No workflow provides" in str(exc)



from OFS.interfaces import IApplication
from plone.distribution.api import distribution as dist_api
from plone.distribution.api import site as site_api
from plone.distribution.core import Distribution
from plone.protect.interfaces import IDisableCSRFProtection
from Products.CMFPlone.browser.admin import AddPloneSite as AddPloneSiteView
from Products.CMFPlone.browser.admin import AppTraverser
from zExceptions import NotFound
from zope.component import adapter
from zope.component import queryMultiAdapter
from zope.interface import alsoProvides
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces import IRequest

import json


@adapter(IApplication, IRequest)
class RestTraverser(AppTraverser):
    """Traverser supporting REST calls in Zope Application root."""

    def publishTraverse(self, request, name):
        if not name.startswith("@"):
            return super().publishTraverse(request, name)

        # This will fail with an AttributeError unless the url has ++api++,
        # otherwise the request has no _rest_service_id.
        # Result is a NotFound error.
        service = queryMultiAdapter(
            (self.context, request), name=request._rest_service_id + name
        )
        if service is not None:
            return service
        # No service, fallback to regular view.
        # But this is unlikely to happen, unless you visit a url like
        # http://localhost:8082/++api++/@ok instead of
        # http://localhost:8082/@@ok
        # Note that a request for http://localhost:8082/++api++/@@ok
        # does not even come in this traverser.
        # Anyway, strip off the '@' sign.
        name = name.lstrip("@")
        view = queryMultiAdapter((self.context, request), name=name)
        if view is not None:
            return view
        raise NotFound(self.context)


class Overview(BrowserView):
    """Overview page for Zope Root."""


class AddPloneSite(AddPloneSiteView):
    """Add Plone site."""

    distribution: Distribution = None
    default_data: dict = None

    def __call__(self):
        context = self.context
        form = self.request.form
        distribution_name = form.get("distribution", "default")
        site_id = form.get("site_id")
        self.distribution = dist_api.get(distribution_name)
        submitted = form.get("form.submitted", False)
        if submitted:
            # CSRF protect. DO NOT use auto CSRF protection for adding a site
            alsoProvides(self.request, IDisableCSRFProtection)
            site = site_api.create(
                context,
                distribution_name,
                profile_id="Products.CMFPlone:plone",
                answers=form,
            )
            self.request.response.redirect(site.absolute_url())
            return ""
        else:
            self.default_data = json.dumps(
                {
                    "site_id": site_id,
                    "default_language": self.browser_language(),
                    "portal_timezone": "UTC",
                },
                indent=2,
            )
        return self.index()



from AccessControl import getSecurityManager
from AccessControl.Permissions import view as View
from datetime import datetime
from datetime import timezone
from plone.base.interfaces import IPloneSiteRoot
from plone.distribution.api import distribution as dist_api
from plone.distribution.core import Distribution
from plone.distribution.core import SiteCreationReport
from plone.distribution.handler import default_handler
from plone.distribution.handler import default_pre_handler
from plone.distribution.utils.validation import validate_answers
from plone.registry.interfaces import IRegistry
from Products.CMFPlone.events import SiteManagerCreatedEvent
from Products.CMFPlone.Portal import PloneSite
from Products.GenericSetup.tool import SetupTool
from typing import List
from ZODB.broken import Broken
from zope.annotation.interfaces import IAnnotations
from zope.component import queryUtility
from zope.component.hooks import setSite
from zope.event import notify
from zope.lifecycleevent import ObjectCreatedEvent

import transaction


_TOOL_ID = "portal_setup"
_DEFAULT_PROFILE = "Products.CMFPlone:plone"


def _handlers_for_distribution(distribution: Distribution):
    """Return pre_handler, handler and post_handler for a distribution."""
    pre_handler = (
        distribution.pre_handler if distribution.pre_handler else default_pre_handler
    )
    handler = distribution.handler if distribution.handler else default_handler
    return pre_handler, handler, distribution.post_handler


def _create_bare_site(context, answers: dict, profile_id: str) -> PloneSite:
    """Create a Plone site."""
    site_id = answers.get("site_id")
    title = answers.get("title")
    description = answers.get("description", "")
    default_language = answers.get("default_language", "en")
    portal_timezone = answers.get("portal_timezone", "UTC")
    # Create the Plone Site
    site = PloneSite(site_id)
    notify(ObjectCreatedEvent(site))
    context[site_id] = site
    site = context[site_id]
    # Set site language
    site.setLanguage(default_language)
    # Register SetupTool (portal_setup)
    site[_TOOL_ID] = SetupTool(_TOOL_ID)
    setup_tool = site[_TOOL_ID]
    notify(SiteManagerCreatedEvent(site))
    setSite(site)
    # Apply base profile
    setup_tool.setBaselineContext(f"profile-{profile_id}")
    setup_tool.runAllImportStepsFromProfile(f"profile-{profile_id}")
    # Set default properties (title, description)
    # Do this before applying extension profiles, so the settings from a
    # properties.xml file are applied and not overwritten by this
    props = {
        "title": title,
        "description": description,
    }
    site.manage_changeProperties(**props)
    # Set initial registry values
    reg = queryUtility(IRegistry, context=site)
    reg["plone.portal_timezone"] = portal_timezone
    reg["plone.available_timezones"] = [portal_timezone]
    reg["plone.default_language"] = default_language
    reg["plone.available_languages"] = [default_language]
    reg["plone.site_title"] = title
    return site


def _add_report_to_site(site: PloneSite, distribution_name: str, answers: dict) -> None:
    """Add a report to the newly created site."""
    # Create a report of a site creation
    annotations = IAnnotations(site)
    report = SiteCreationReport(
        distribution_name, datetime.now(tz=timezone.utc), answers
    )
    annotations[dist_api.SITE_REPORT_ANNO] = report


def _create_site(
    context,
    distribution_name: str,
    answers: dict,
    profile_id: str = _DEFAULT_PROFILE,
) -> PloneSite:
    """Create site"""
    distribution = dist_api.get(distribution_name)
    pre_handler, handler, post_handler = _handlers_for_distribution(distribution)
    # Process answers
    answers = pre_handler(answers)
    # Create base site
    site = _create_bare_site(context, answers, profile_id)
    # Run the Distribution handler
    site = handler(distribution, site, answers)
    # Run the Distribution post_handler
    if post_handler:
        site = post_handler(distribution, site, answers)
    # Create a report of a site creation
    _add_report_to_site(site, distribution_name, answers)
    return site


def get_sites(context=None) -> List[PloneSite]:
    """Get all Plone sites.

    :param context: Base context to search for Plone Sites.
    :raises:
        :class:`ValueError`,
    :returns: A list of Plone Sites

    :Example: :ref:`api-site-get_sites-example`
    """
    if not context:
        raise ValueError("Need to provide application root")
    result = []
    secman = getSecurityManager()
    candidates = (obj for obj in context.values() if not isinstance(obj, Broken))
    for obj in candidates:
        if obj.meta_type == "Folder":
            result.extend(get_sites(context=obj))
        elif IPloneSiteRoot.providedBy(obj):
            if secman.checkPermission(View, obj):
                result.append(obj)
        elif obj.getId() in getattr(context, "_mount_points", {}):
            result.extend(get_sites(context=obj))
    return result


def create(
    context,
    distribution_name: str,
    answers: dict,
    profile_id: str = _DEFAULT_PROFILE,
) -> PloneSite:
    """Create a new Plone site using one of the distributions.

    :param context: Context where the site will be created.
    :param distribution_name: Name of distribution to be used.
    :param answers: Payload for site creation.
    :param profile_id: Base profile to be used.
                       default: `Products.CMFPlone:plone`
    :raises:
        :class:`ValueError`,
        :class:`KeyError`,
    :returns: Created Plone Site

    :Example: :ref:`api-site-create-example`
    """
    distribution = dist_api.get(distribution_name)
    # Validate answers
    schema = distribution.schema
    if not validate_answers(answers=answers, schema=schema):
        raise ValueError("Provided answers are not valid")
    with transaction.manager as tm:
        site = _create_site(context, distribution_name, answers, profile_id)
        tm.note(f"Plone site {site.getId()} created with {distribution_name}")
    return site



from jsonschema import exceptions
from jsonschema import validate
from plone.distribution import logger


def validate_answers(answers: dict, schema: dict) -> bool:
    """Validate answers against the provided schema.

    :param answers: Payload for site creation.
    :param schema: A JSON Schema.

    :returns: Boolean indicating if the answers are valid or not.
    """
    is_valid = False
    try:
        validate(instance=answers, schema=schema)
    except exceptions.ValidationError as exc:
        logger.exception("Validation error", exc_info=exc)
    except (exceptions.SchemaError, exceptions.RefResolutionError) as exc:
        logger.exception("Schema error", exc_info=exc)
    else:
        site_id = answers.get("site_id").strip()
        title = answers.get("title").strip()
        is_valid = True if site_id and title else False
    return is_valid



import pytest


@pytest.fixture()
def app(functional):
    return functional["app"]


@pytest.fixture()
def portal_default(functional):
    return functional["portal"]


@pytest.fixture()
def http_request(functional):
    return functional["request"]



from plone import api
from plone.distribution.api.distribution import get_current_distribution
from plone.exportimport.exporters import get_exporter
from plone.exportimport.utils import cli as cli_helpers

import argparse
import sys


CLI_SPEC = {
    "exporter": {
        "description": "Export Plone Site content",
        "options": {
            "zopeconf": "Path to zope.conf",
            "site": "Plone site ID to export the content from",
            "--include-revisions": "Include revision history",
        },
    },
}


def _parse_args(description: str, options: dict, args: list):
    parser = argparse.ArgumentParser(description=description)
    for key, help in options.items():
        if key.startswith("-"):
            parser.add_argument(key, action="store_true", help=help)
        else:
            parser.add_argument(key, help=help)
    namespace, _ = parser.parse_known_args(args[1:])
    return namespace


def export(args=sys.argv):
    """Export a Plone site to a distribution."""
    logger = cli_helpers.get_logger("Exporter")
    exporter_cli = CLI_SPEC["exporter"]
    # We get an argparse.Namespace instance.
    namespace = _parse_args(exporter_cli["description"], exporter_cli["options"], args)
    app = cli_helpers.get_app(namespace.zopeconf)
    site = cli_helpers.get_site(app, namespace.site, logger)
    distribution = get_current_distribution(site)
    path = distribution.directory / "content"
    if not path:
        logger.error(f"{namespace.path} does not exist, please create it first.")
        sys.exit(1)
    logger.info(f"Exporting Plone site at /{site.id}")
    logger.info(f" Target path: {path}")
    with api.env.adopt_roles(["Manager"]):
        results = get_exporter(site).export_site(path, options=namespace)
    for item in results[1:]:
        logger.info(f" Wrote {item.relative_to(path)}")



from .layer import PloneDistributionFixture
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneSandboxLayer
from plone.app.testing.interfaces import PLONE_SITE_ID
from plone.testing.zope import WSGI_SERVER_FIXTURE


DISTRIBUTION = "testing"
ANSWERS = {
    "site_id": PLONE_SITE_ID,
    "title": "Testing Plone Site",
    "description": "A Plone Site",
    "site_logo": "name=teste;data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg==",  # noQA
    "default_language": "en",
    "portal_timezone": "America/Sao_Paulo",
    "setup_content": True,
}


class BaseFixture(PloneDistributionFixture):
    PACKAGE_NAME: str = "plone.distribution.testing"
    SITES = ((DISTRIBUTION, ANSWERS),)


BASE_FIXTURE = BaseFixture()


class Layer(PloneSandboxLayer):
    defaultBases = (BASE_FIXTURE,)


FIXTURE = Layer()


INTEGRATION_TESTING = IntegrationTesting(
    bases=(FIXTURE,),
    name="DistributionLayer:IntegrationTesting",
)

FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(FIXTURE, WSGI_SERVER_FIXTURE),
    name="DistributionLayer:FunctionalTesting",
)



from pathlib import Path
from plone import api
from plone.distribution.core import Distribution
from plone.distribution.core import SiteCreationReport
from plone.distribution.registry import _distribution_registry
from plone.distribution.registry import DistributionRegistry
from Products.CMFPlone.Portal import PloneSite
from typing import List
from typing import Optional
from typing import Union
from zope.annotation.interfaces import IAnnotations

import os


base_folder = Path(__file__).parent.parent / "distributions"
SITE_REPORT_ANNO = "__plone_distribution_report__"


def _allow_list() -> List[str]:
    """Return a list of allowed distributions."""
    allowed_distributions = os.environ.get("ALLOWED_DISTRIBUTIONS", "")
    if allowed_distributions:
        allowed_distributions = allowed_distributions.split(",")
    return allowed_distributions


def get_registry() -> DistributionRegistry:
    """Return the Distribution Registry.

    :returns: Distribution Registry instance.

    :Example: :ref:`api-distribution-get_registry-example`
    """
    return _distribution_registry


def get_distributions(filter: bool = True) -> List[Distribution]:
    """Get available Plone distributions.

    :param filter: Return only registered Distributions that
                   are also listed on **ALLOWED_DISTRIBUTIONS**
                   environment variable.
    :returns: List of registered distributions.

    :Example: :ref:`api-distribution-get_distributions-example`
    """
    registry = get_registry()
    all_distributions = registry.enumerate_distributions()
    allowed_distributions = _allow_list()
    if filter and allowed_distributions:
        return [d for d in all_distributions if d.name in allowed_distributions]
    return all_distributions


def get(name: str) -> Distribution:
    """Get a distribution.

    :param name: Distribution name.
    :raises:
        :class:`ValueError`,
    :returns: A Plone Distribution

    :Example: :ref:`api-distribution-get-example`
    """
    registry = get_registry()
    distribution = registry.lookup(name)
    if not distribution:
        raise ValueError(f"No distribution named {name}")
    return distribution


def get_creation_report(site: PloneSite) -> Union[SiteCreationReport, None]:
    """Return a site creation report for a Plone site.

    :param site: Plone Site.
    :returns: SiteCreationReport with distribution name, creation date and
              answers used to create the site.

    :Example: :ref:`api-distribution-get_creation_report-example`
    """
    annotations = IAnnotations(site)
    return annotations.get(SITE_REPORT_ANNO, None)


def get_current_distribution(site: PloneSite = None) -> Optional[Distribution]:
    """Get the distribution used to create the current site."""
    if site is None:
        site = api.portal.get()
    creation_report = get_creation_report(site)
    dist = None
    if creation_report:
        name = creation_report.name or creation_report.answers.get("distribution", "")
        dist = get(name=name)
    return dist



from importlib.metadata import distribution
from plone.distribution.api import distribution as dist_api
from plone.restapi.services.system.get import SystemGet as BaseService


plone_distribution_version = distribution("plone.distribution").version


class SystemGet(BaseService):
    def distribution_info(self) -> str:
        """Get distribution information if site was created with a distribution."""
        info = ""
        portal = self.context
        report = dist_api.get_creation_report(portal)
        if report:
            dist_name = report.name
            try:
                dist_title = dist_api.get(dist_name).title
            except ValueError:
                dist_title = "Uninstalled"
            info = f"{dist_title} ({dist_name})"
        return info

    def reply(self):
        response = super().reply()
        response["plone.distribution"] = plone_distribution_version
        distribution_info = self.distribution_info()
        if distribution_info:
            response["distribution"] = distribution_info
        return response



from plone import api
from plone.distribution.api import distribution as dist_api
from plone.distribution.api import site as site_api
from Products.CMFPlone.Portal import PloneSite

import pytest


@pytest.fixture
def answers() -> dict:
    """Answers payload for new site creation."""
    return {
        "site_id": "PloneBrasil",
        "title": "Plone site in Brazil",
        "description": "A simple Plone Site",
        "default_language": "pt-br",
        "portal_timezone": "America/Sao_Paulo",
        "setup_content": False,  # Avoid committing a transaction
    }


@pytest.fixture
def site(app, answers) -> PloneSite:
    """Create a new Plone site via a distribution."""
    site_id = answers["site_id"]
    distribution_name = "testing"
    yield site_api.create(app, distribution_name, answers)
    app.manage_delObjects([site_id])


class TestApiSite:
    def test_get_sites(self, app, integration):
        sites = site_api.get_sites(app)
        # Integration test creates a Plone Site
        assert len(sites) == 1
        site = sites[0]
        assert isinstance(site, PloneSite)

    @pytest.mark.parametrize(
        "distribution_name",
        [
            "testing",
        ],
    )
    def test_create(self, app, integration, answers, distribution_name):
        sites = site_api.get_sites(app)
        total_sites = len(sites)
        with api.env.adopt_roles(["Manager"]):
            answers["site_id"] = f"{distribution_name}_01"
            new_site = site_api.create(app, distribution_name, answers)
        sites = site_api.get_sites(app)
        assert len(sites) == total_sites + 1
        site = sites[-1]
        assert isinstance(site, PloneSite)
        assert site.title == new_site.title

    def test_get_creation_report_new_site(self, site):
        from datetime import datetime
        from plone.distribution.core import SiteCreationReport

        report = dist_api.get_creation_report(site)
        assert isinstance(report, SiteCreationReport)
        assert report.name == "testing"
        assert isinstance(report.date, datetime)
        assert isinstance(report.answers, dict)
        assert report.answers["title"] == site.title



from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.distribution.api.distribution import SITE_REPORT_ANNO
from plone.restapi.testing import RelativeSession
from zope.annotation.interfaces import IAnnotations

import pytest


@pytest.fixture
def portal_uninstalled(portal):
    """Mock the uninstallation of a distribution."""
    import transaction

    annotations = IAnnotations(portal)
    report = annotations[SITE_REPORT_ANNO]
    current = report.name
    with transaction.manager:
        report.name = "mocked-dist"
        annotations[SITE_REPORT_ANNO] = report
    yield portal
    with transaction.manager:
        report.name = current
        annotations[SITE_REPORT_ANNO] = report


class TestServicesSystemGET:
    @pytest.fixture(autouse=True)
    def _api_session(self, portal):
        url = portal.absolute_url()
        api_session = RelativeSession(url)
        api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        api_session.headers.update({"Accept": "application/json"})
        self.api_session = api_session

    def test_system_get(self, portal):
        response = self.api_session.get("@system")
        assert response.status_code == 200
        data = response.json()
        assert isinstance(data, dict)
        assert "plone.distribution" in data

    @pytest.mark.parametrize(
        "attr,expected",
        [
            ("distribution", "Testing Plone Site (testing)"),
        ],
    )
    def test_system_get_distribution(self, portal, attr, expected):
        response = self.api_session.get("@system")
        data = response.json()
        assert attr in data
        assert data[attr] == expected


class TestServicesSystemGETUninstalled:
    @pytest.fixture(autouse=True)
    def _api_session(self, portal_uninstalled):
        url = portal_uninstalled.absolute_url()
        api_session = RelativeSession(url)
        api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        api_session.headers.update({"Accept": "application/json"})
        self.api_session = api_session

    def test_system_get(self, portal_uninstalled):
        response = self.api_session.get("@system")
        assert response.status_code == 200
        data = response.json()
        assert isinstance(data, dict)
        assert "plone.distribution" in data
        assert "distribution" in data
        assert data["distribution"] == "Uninstalled (mocked-dist)"
